# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import itertools
import textwrap
import time
from typing import TYPE_CHECKING
from typing import Generic
from typing import TypeVar
from typing import Union
from typing import overload

from ._hangcheck import HangCheckFactory
from ._notifier import Notifier

if TYPE_CHECKING:
    from typing import Any

    from twisted.internet.interfaces import IReactorTime

    _T = TypeVar('_T')

StrOrBytesType = TypeVar("StrOrBytesType", bound=Union[str, bytes])


__all__ = [
    "HangCheckFactory",
    "Notifier",
    "Obfuscated",
    "now",
    "remove_userpassword",
    "rewrap",
]


def remove_userpassword(url: str) -> str:
    if '@' not in url:
        return url
    if '://' not in url:
        return url

    # urlparse would've been nice, but doesn't support ssh... sigh
    (protocol, repo_url) = url.split('://')
    repo_url = repo_url.split('@')[-1]

    return protocol + '://' + repo_url


def now(_reactor: IReactorTime | None = None) -> float:
    if _reactor is not None and hasattr(_reactor, "seconds"):
        return _reactor.seconds()
    return time.time()


class Obfuscated(Generic[StrOrBytesType]):
    """An obfuscated string in a command"""

    def __init__(self, real: StrOrBytesType, fake: StrOrBytesType) -> None:
        self.real = real
        self.fake = fake

    def __str__(self) -> str:
        return str(self.fake)

    def __repr__(self) -> str:
        return repr(self.fake)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            other.__class__ is self.__class__
            and other.real == self.real
            and other.fake == self.fake
        )

    @overload
    @staticmethod
    def to_text(s: bytes) -> bytes: ...

    @overload
    @staticmethod
    def to_text(s: Any) -> str: ...

    @staticmethod
    def to_text(s: Any) -> str | bytes:
        if isinstance(s, (str, bytes)):
            return s
        return str(s)

    @overload
    @staticmethod
    def get_real(command: list[Obfuscated | _T]) -> list[str | _T]: ...

    @overload
    @staticmethod
    def get_real(command: _T) -> _T: ...

    @staticmethod
    def get_real(command: list[Obfuscated | Any] | Any) -> list[Any] | Any:
        rv = command
        if isinstance(command, list):
            rv = []
            for elt in command:
                if isinstance(elt, Obfuscated):
                    rv.append(elt.real)
                else:
                    rv.append(Obfuscated.to_text(elt))
        return rv

    @overload
    @staticmethod
    def get_fake(command: list[Obfuscated | _T]) -> list[str | _T]: ...

    @overload
    @staticmethod
    def get_fake(command: _T) -> _T: ...

    @staticmethod
    def get_fake(command: list[Obfuscated | Any] | Any) -> list[Any] | Any:
        rv = command
        if isinstance(command, list):
            rv = []
            for elt in command:
                if isinstance(elt, Obfuscated):
                    rv.append(elt.fake)
                else:
                    rv.append(Obfuscated.to_text(elt))
        return rv


def rewrap(text: str, width: int | None = None) -> str:
    """
    Rewrap text for output to the console.

    Removes common indentation and rewraps paragraphs according to the console
    width.

    Line feeds between paragraphs preserved.
    Formatting of paragraphs that starts with additional indentation
    preserved.
    """

    if width is None:
        width = 80

    # Remove common indentation.
    text = textwrap.dedent(text)

    def needs_wrapping(line: str) -> bool:
        # Line always non-empty.
        return not line[0].isspace()

    # Split text by lines and group lines that comprise paragraphs.
    wrapped_text = ""
    for do_wrap, lines in itertools.groupby(text.splitlines(True), key=needs_wrapping):
        paragraph = ''.join(lines)

        if do_wrap:
            paragraph = textwrap.fill(paragraph, width)

        wrapped_text += paragraph

    return wrapped_text


def twisted_connection_string_to_ws_url(description: str) -> str:
    from twisted.internet.endpoints import _parse

    args, kwargs = _parse(description)
    protocol = args.pop(0).upper()

    host = kwargs.get('host', None)
    port = kwargs.get('port', None)

    if protocol == 'TCP':
        port = kwargs.get('port', 80)

        if len(args) == 2:
            host = args[0]
            port = args[1]
        elif len(args) == 1:
            if "host" in kwargs:
                host = kwargs['host']
                port = args[0]
            else:
                host = args[0]
                port = kwargs.get('port', port)

    if host is None or host == '' or port is None:
        raise ValueError('Host and port must be specified in connection string')

    return f"ws://{host}:{port}"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import platform
from unittest.case import SkipTest
from urllib import request as urllib_request

from twisted.internet import reactor
from twisted.python.filepath import FilePath
from twisted.trial import unittest

import buildbot.buildbot_net_usage_data
from buildbot import config
from buildbot.buildbot_net_usage_data import _sendBuildbotNetUsageData
from buildbot.buildbot_net_usage_data import computeUsageData
from buildbot.buildbot_net_usage_data import linux_distribution
from buildbot.config import BuilderConfig
from buildbot.master import BuildMaster
from buildbot.plugins import steps
from buildbot.process.factory import BuildFactory
from buildbot.schedulers.forcesched import ForceScheduler
from buildbot.test.util.integration import DictLoader
from buildbot.test.util.warnings import assertProducesWarning
from buildbot.warnings import ConfigWarning
from buildbot.worker.base import Worker


class Tests(unittest.TestCase):
    def getMaster(self, config_dict):
        """
        Create a started ``BuildMaster`` with the given configuration.
        """
        basedir = FilePath(self.mktemp())
        basedir.createDirectory()
        master = BuildMaster(basedir.path, reactor=reactor, config_loader=DictLoader(config_dict))
        master.config = master.config_loader.loadConfig()
        return master

    def getBaseConfig(self):
        return {
            'builders': [
                BuilderConfig(
                    name="testy",
                    workernames=["local1", "local2"],
                    factory=BuildFactory([steps.ShellCommand(command='echo hello')]),
                ),
            ],
            'workers': [Worker('local' + str(i), 'pass') for i in range(3)],
            'schedulers': [ForceScheduler(name="force", builderNames=["testy"])],
            'protocols': {'null': {}},
            'multiMaster': True,
        }

    def test_basic(self):
        self.patch(config.master, "get_is_in_unit_tests", lambda: False)
        with assertProducesWarning(
            ConfigWarning,
            message_pattern=r"`buildbotNetUsageData` is not configured and defaults to basic.",
        ):
            master = self.getMaster(self.getBaseConfig())
        data = computeUsageData(master)
        self.assertEqual(
            sorted(data.keys()),
            sorted(['versions', 'db', 'platform', 'installid', 'mq', 'plugins', 'www_plugins']),
        )
        self.assertEqual(data['plugins']['buildbot/worker/base/Worker'], 3)
        self.assertEqual(
            sorted(data['plugins'].keys()),
            sorted([
                'buildbot/schedulers/forcesched/ForceScheduler',
                'buildbot/worker/base/Worker',
                'buildbot/steps/shell/ShellCommand',
                'buildbot/config/builder/BuilderConfig',
            ]),
        )

    def test_full(self):
        c = self.getBaseConfig()
        c['buildbotNetUsageData'] = 'full'
        master = self.getMaster(c)
        data = computeUsageData(master)
        self.assertEqual(
            sorted(data.keys()),
            sorted([
                'versions',
                'db',
                'installid',
                'platform',
                'mq',
                'plugins',
                'builders',
                'www_plugins',
            ]),
        )

    def test_custom(self):
        c = self.getBaseConfig()

        def myCompute(data):
            return {"db": data['db']}

        c['buildbotNetUsageData'] = myCompute
        master = self.getMaster(c)
        data = computeUsageData(master)
        self.assertEqual(sorted(data.keys()), sorted(['db']))

    def test_urllib(self):
        self.patch(buildbot.buildbot_net_usage_data, '_sendWithRequests', lambda _, __: None)

        class FakeRequest:
            def __init__(self, *args, **kwargs):
                self.args = args
                self.kwargs = kwargs

        open_url = []

        class urlopen:
            def __init__(self, r):
                self.request = r
                open_url.append(self)

            def read(self):
                return "ok"

            def close(self):
                pass

        self.patch(urllib_request, "Request", FakeRequest)
        self.patch(urllib_request, "urlopen", urlopen)
        _sendBuildbotNetUsageData({'foo': 'bar'})
        self.assertEqual(len(open_url), 1)
        self.assertEqual(
            open_url[0].request.args,
            (
                'https://events.buildbot.net/events/phone_home',
                b'{"foo": "bar"}',
                {'Content-Length': '14', 'Content-Type': 'application/json'},
            ),
        )

    def test_real(self):
        if "TEST_BUILDBOTNET_USAGEDATA" not in os.environ:
            raise SkipTest(
                "_sendBuildbotNetUsageData real test only run when environment variable"
                " TEST_BUILDBOTNET_USAGEDATA is set"
            )

        _sendBuildbotNetUsageData({'foo': 'bar'})

    def test_linux_distro(self):
        system = platform.system()
        if system != "Linux":
            raise SkipTest("test is only for linux")
        distro = linux_distribution()
        self.assertEqual(len(distro), 2)
        self.assertNotIn("unknown", distro[0])
        # Rolling distributions like Arch Linux (arch) does not have VERSION_ID
        if distro[0] not in ["arch", "gentoo", "antergos"]:
            self.assertNotIn("unknown", distro[1])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import os
import re
import stat
from pathlib import Path
from pathlib import PurePath
from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Sequence

from packaging.version import parse as parse_version
from twisted.internet import defer
from twisted.python import log

from buildbot import config
from buildbot.process import buildstep
from buildbot.process import remotecommand
from buildbot.process.properties import Properties
from buildbot.steps.worker import CompositeStepMixin
from buildbot.util import ComparableMixin
from buildbot.util import bytes2unicode
from buildbot.util.misc import writeLocalFile
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    from buildbot.changes.gitpoller import GitPoller
    from buildbot.interfaces import IRenderable
    from buildbot.util.git_credential import GitCredentialOptions

RC_SUCCESS = 0


def getSshArgsForKeys(keyPath, knownHostsPath):
    args = ['-o', 'BatchMode=yes']
    if keyPath is not None:
        args += ['-i', keyPath]
    if knownHostsPath is not None:
        args += ['-o', f'UserKnownHostsFile={knownHostsPath}']
    return args


def escapeShellArgIfNeeded(arg):
    if re.match(r"^[a-zA-Z0-9_-]+$", arg):
        return arg
    return f'"{arg}"'


def getSshCommand(keyPath, knownHostsPath):
    command = ['ssh', *getSshArgsForKeys(keyPath, knownHostsPath)]
    command = [escapeShellArgIfNeeded(arg) for arg in command]
    return ' '.join(command)


def scp_style_to_url_syntax(address, port=22, scheme='ssh'):
    if any(['://' in address, ':\\' in address, ':' not in address]):
        # the address already has a URL syntax or is a local path
        return address
    host, path = address.split(':')
    return f'{scheme}://{host}:{port}/{path}'


def check_ssh_config(
    logname: str,
    ssh_private_key: IRenderable | str | None,
    ssh_host_key: IRenderable | str | None,
    ssh_known_hosts: IRenderable | str | None,
):
    if ssh_host_key is not None and ssh_private_key is None:
        config.error(f'{logname}: sshPrivateKey must be provided in order use sshHostKey')

    if ssh_known_hosts is not None and ssh_private_key is None:
        config.error(f'{logname}: sshPrivateKey must be provided in order use sshKnownHosts')

    if ssh_host_key is not None and ssh_known_hosts is not None:
        config.error(f'{logname}: only one of sshKnownHosts and sshHostKey can be provided')


class GitMixin:
    def setupGit(self):
        self.gitInstalled = False
        self.supportsBranch = False
        self.supportsProgress = False
        self.supportsSubmoduleForce = False
        self.supportsSubmoduleCheckout = False
        self.supportsSshPrivateKeyAsEnvOption = False
        self.supportsSshPrivateKeyAsConfigOption = False
        self.supportsFilters = False
        self.supports_lsremote_symref = False
        self.supports_credential_store = False

    def parseGitFeatures(self, version_stdout):
        match = re.match(r"^git version (\d+(\.\d+)*)", version_stdout)
        if not match:
            return

        version = parse_version(match.group(1))

        self.gitInstalled = True
        if version >= parse_version("1.6.5"):
            self.supportsBranch = True
        if version >= parse_version("1.7.2"):
            self.supportsProgress = True
        if version >= parse_version("1.7.6"):
            self.supportsSubmoduleForce = True
        if version >= parse_version("1.7.8"):
            self.supportsSubmoduleCheckout = True
        if version >= parse_version("1.7.9"):
            self.supports_credential_store = True
        if version >= parse_version("2.3.0"):
            self.supportsSshPrivateKeyAsEnvOption = True
        if version >= parse_version("2.8.0"):
            # https://github.com/git/git/blob/v2.8.0/Documentation/RelNotes/2.8.0.txt#L72-L73
            self.supports_lsremote_symref = True
        if version >= parse_version("2.10.0"):
            self.supportsSshPrivateKeyAsConfigOption = True
        if version >= parse_version("2.27.0"):
            self.supportsFilters = True

    def adjustCommandParamsForSshPrivateKey(
        self, command, env, keyPath, sshWrapperPath=None, knownHostsPath=None
    ):
        ssh_command = getSshCommand(keyPath, knownHostsPath)

        if self.supportsSshPrivateKeyAsConfigOption:
            command.append('-c')
            command.append(f'core.sshCommand={ssh_command}')
        elif self.supportsSshPrivateKeyAsEnvOption:
            env['GIT_SSH_COMMAND'] = ssh_command
        else:
            if sshWrapperPath is None:
                raise RuntimeError('Only SSH wrapper script is supported but path not given')
            env['GIT_SSH'] = sshWrapperPath


def getSshWrapperScriptContents(keyPath, knownHostsPath=None):
    ssh_command = getSshCommand(keyPath, knownHostsPath)

    # note that this works on windows if using git with MINGW embedded.
    return f'#!/bin/sh\n{ssh_command} "$@"\n'


def getSshKnownHostsContents(hostKey: str) -> str:
    host_name = '*'
    return f'{host_name} {hostKey}'


def ensureSshKeyNewline(privateKey: str) -> str:
    """Ensure key has trailing newline

    Providers can be configured to strip newlines from secrets. This feature
    breaks SSH key use within the Git module. This helper function ensures that
    when an ssh key is provided for a git step that is contains the trailing
    newline.
    """
    if privateKey.endswith("\n") or privateKey.endswith("\r") or privateKey.endswith("\r\n"):
        return privateKey
    return privateKey + "\n"


class GitStepMixin(GitMixin):
    _git_auth: GitStepAuth

    def setupGitStep(self):
        self.setupGit()

        if not self.repourl:
            config.error("Git: must provide repourl.")

        if not hasattr(self, '_git_auth'):
            self._git_auth = GitStepAuth(self)

    def setup_repourl(self):
        # Use standard URL syntax to enable the use of a dedicated SSH port
        self.repourl = scp_style_to_url_syntax(self.repourl, self.port)

    def setup_git_auth(
        self,
        ssh_private_key: IRenderable | str | None,
        ssh_host_key: IRenderable | str | None,
        ssh_known_hosts: IRenderable | str | None,
        git_credential_options: GitCredentialOptions | None = None,
    ) -> None:
        self._git_auth = GitStepAuth(
            self,
            ssh_private_key,
            ssh_host_key,
            ssh_known_hosts,
            git_credential_options=git_credential_options,
        )

    def _get_auth_data_workdir(self) -> str:
        raise NotImplementedError()

    @defer.inlineCallbacks
    def _dovccmd(self, command, abandonOnFailure=True, collectStdout=False, initialStdin=None):
        full_command = ['git']
        full_env = self.env.copy() if self.env else {}

        if self.config is not None:
            for name, value in self.config.items():
                full_command.append('-c')
                full_command.append(f'{name}={value}')

        if command and self._git_auth.is_auth_needed_for_git_command(command[0]):
            self._git_auth.adjust_git_command_params_for_auth(
                full_command,
                full_env,
                self._get_auth_data_workdir(),
                self,
            )

        full_command.extend(command)

        # check for the interruptSignal flag
        sigtermTime = None
        interruptSignal = None

        # If possible prefer to send a SIGTERM to git before we send a SIGKILL.
        # If we send a SIGKILL, git is prone to leaving around stale lockfiles.
        # By priming it with a SIGTERM first we can ensure that it has a chance to shut-down
        # gracefully before getting terminated
        if not self.workerVersionIsOlderThan("shell", "2.16"):
            # git should shut-down quickly on SIGTERM.  If it doesn't don't let it
            # stick around for too long because this is on top of any timeout
            # we have hit.
            sigtermTime = 1
        else:
            # Since sigtermTime is unavailable try to just use SIGTERM by itself instead of
            # killing.  This should be safe.
            if self.workerVersionIsOlderThan("shell", "2.15"):
                log.msg(
                    "NOTE: worker does not allow master to specify "
                    "interruptSignal. This may leave a stale lockfile around "
                    "if the command is interrupted/times out\n"
                )
            else:
                interruptSignal = 'TERM'

        cmd = remotecommand.RemoteShellCommand(
            self.workdir,
            full_command,
            env=full_env,
            logEnviron=self.logEnviron,
            timeout=self.timeout,
            sigtermTime=sigtermTime,
            interruptSignal=interruptSignal,
            collectStdout=collectStdout,
            initialStdin=initialStdin,
        )
        cmd.useLog(self.stdio_log, False)
        yield self.runCommand(cmd)

        if abandonOnFailure and cmd.didFail():
            log.msg(f"Source step failed while running command {cmd}")
            raise buildstep.BuildStepFailed()
        if collectStdout:
            return cmd.stdout
        return cmd.rc

    @defer.inlineCallbacks
    def checkFeatureSupport(self):
        stdout = yield self._dovccmd(['--version'], collectStdout=True)

        self.parseGitFeatures(stdout)

        return self.gitInstalled


class AbstractGitAuth(ComparableMixin):
    compare_attrs: ClassVar[Sequence[str]] = (
        "ssh_private_key",
        "ssh_host_key",
        "ssh_known_hosts",
        "git_credential_options",
    )

    def __init__(
        self,
        ssh_private_key: IRenderable | str | None = None,
        ssh_host_key: IRenderable | str | None = None,
        ssh_known_hosts: IRenderable | str | None = None,
        git_credential_options: GitCredentialOptions | None = None,
    ) -> None:
        self.did_download_auth_files = False

        self.ssh_private_key = ssh_private_key
        self.ssh_host_key = ssh_host_key
        self.ssh_known_hosts = ssh_known_hosts

        self.git_credential_options = git_credential_options

        check_ssh_config('Git', self.ssh_private_key, self.ssh_host_key, self.ssh_known_hosts)

    @property
    def is_auth_needed(self) -> bool:
        return self.ssh_private_key is not None or self.git_credential_options is not None

    def is_auth_needed_for_git_command(self, git_command: str) -> bool:
        if not git_command:
            return False

        if not self.is_auth_needed:
            return False

        git_commands_that_need_auth = [
            'clone',
            'credential',
            'checkout',
            'fetch',
            'ls-remote',
            'push',
            'submodule',
        ]
        if git_command in git_commands_that_need_auth:
            return True
        return False

    @property
    def _path_module(self):
        raise NotImplementedError()

    @property
    def _master(self):
        raise NotImplementedError()

    def _get_ssh_private_key_path(self, ssh_data_path: str) -> str:
        return self._path_module.join(ssh_data_path, 'ssh-key')

    def _get_ssh_host_key_path(self, ssh_data_path: str) -> str:
        return self._path_module.join(ssh_data_path, 'ssh-known-hosts')

    def _get_ssh_wrapper_script_path(self, ssh_data_path: str) -> str:
        return self._path_module.join(ssh_data_path, 'ssh-wrapper.sh')

    def _get_credential_store_file_path(self, ssh_data_path):
        return self._path_module.join(ssh_data_path, '.git-credentials')

    def _adjust_command_params_for_ssh_private_key(
        self,
        full_command: list[str],
        full_env: dict[str, str],
        workdir: str,
        git_mixin: GitMixin,
    ) -> None:
        if self.ssh_private_key is None:
            return

        key_path = self._get_ssh_private_key_path(workdir)
        host_key_path = None
        if self.ssh_host_key is not None or self.ssh_known_hosts is not None:
            host_key_path = self._get_ssh_host_key_path(workdir)

        ssh_wrapper_path = self._get_ssh_wrapper_script_path(workdir)

        git_mixin.adjustCommandParamsForSshPrivateKey(
            full_command,
            full_env,
            key_path,
            ssh_wrapper_path,
            host_key_path,
        )

    def _adjust_command_params_for_credential_store(
        self,
        full_command: list[str],
        workdir: str,
        git_mixin: GitMixin,
    ):
        if self.git_credential_options is None:
            return

        if not git_mixin.supports_credential_store:
            raise RuntimeError('git credential-store is not supported')

        credentials_path = self._get_credential_store_file_path(workdir)

        # This will unset the `credential.helper` config for this command
        # so that system/global credential store is not used
        # NOTE: This could be optional allowing credential retrieval from system sources
        # However, it would need the store process (`credential approve`) to pass it
        # as `credential approve` will store the credential in ALL credential helpers
        full_command.extend([
            '-c',
            'credential.helper=',
        ])

        full_command.extend([
            '-c',
            f'credential.helper=store "--file={credentials_path}"',
        ])

        if self.git_credential_options.use_http_path is not None:
            # Whether or not to only use domain for credential lookup
            value = 'true' if self.git_credential_options.use_http_path else 'false'
            full_command.extend([
                '-c',
                f'credential.useHttpPath={value}',
            ])

    def adjust_git_command_params_for_auth(
        self,
        full_command: list[str],
        full_env: dict[str, str],
        workdir: str,
        git_mixin: GitMixin,
    ) -> None:
        self._adjust_command_params_for_ssh_private_key(
            full_command,
            full_env,
            workdir=workdir,
            git_mixin=git_mixin,
        )
        self._adjust_command_params_for_credential_store(
            full_command,
            workdir=workdir,
            git_mixin=git_mixin,
        )

    @async_to_deferred
    async def _dovccmd(
        self,
        command: list[str],
        initial_stdin: str | None = None,
        workdir: str | None = None,
    ) -> None:
        raise NotImplementedError()

    async def _download_file(
        self,
        path: str,
        content: str,
        mode: int,
        workdir: str | None = None,
    ) -> None:
        raise NotImplementedError()

    @async_to_deferred
    async def _download_ssh_files(
        self,
        private_key: str,
        host_key: str | None,
        known_hosts: str | None,
        workdir: str,
        download_wrapper_script: bool = False,
    ) -> None:
        private_key_path = self._get_ssh_private_key_path(workdir)
        private_key = ensureSshKeyNewline(private_key)
        await self._download_file(
            private_key_path,
            private_key,
            mode=stat.S_IRUSR,
            workdir=workdir,
        )

        known_hosts_path = self._get_ssh_host_key_path(workdir)
        known_hosts_contents = None
        if known_hosts is not None:
            known_hosts_contents = known_hosts
        elif host_key is not None:
            known_hosts_contents = getSshKnownHostsContents(host_key)

        if known_hosts_contents is not None:
            await self._download_file(
                known_hosts_path,
                known_hosts_contents,
                mode=stat.S_IRUSR,
                workdir=workdir,
            )

        if download_wrapper_script:
            script_path = self._get_ssh_wrapper_script_path(workdir)
            script_contents = getSshWrapperScriptContents(
                private_key_path,
                (known_hosts_path if known_hosts_contents is not None else None),
            )

            await self._download_file(
                script_path,
                script_contents,
                mode=stat.S_IRWXU,
                workdir=workdir,
            )

    @async_to_deferred
    async def _download_credentials(
        self,
        credentials: list[str],
        workdir: str,
    ) -> None:
        for creds in credentials:
            # Using credential approve here instead of directly writing to the file
            # as recommended by Git doc (https://git-scm.com/docs/git-credential-store#_storage_format)
            # "Do not view or edit the file with editors."
            await self._dovccmd(
                ['credential', 'approve'],
                initial_stdin=creds,
                workdir=workdir,
            )

    @async_to_deferred
    async def download_auth_files_if_needed(
        self,
        workdir: str,
        download_wrapper_script: bool = False,
    ) -> int:
        p = Properties()
        p.master = self._master

        private_key: str | None = await p.render(self.ssh_private_key)
        host_key: str | None = await p.render(self.ssh_host_key)
        known_hosts: str | None = await p.render(self.ssh_known_hosts)

        if private_key is not None:
            await self._download_ssh_files(
                private_key,
                host_key,
                known_hosts,
                workdir,
                download_wrapper_script,
            )
            self.did_download_auth_files = True

        if self.git_credential_options is not None:
            credentials: list[str] = []
            for creds in self.git_credential_options.credentials:
                rendered: str | None = await p.render(creds)
                if rendered:
                    credentials.append(rendered)

            if credentials:
                await self._download_credentials(credentials, workdir)
                self.did_download_auth_files = True

        return RC_SUCCESS

    @async_to_deferred
    async def remove_auth_files_if_needed(self, workdir: str) -> int:
        raise NotImplementedError()


class GitStepAuth(AbstractGitAuth):
    def __init__(
        self,
        # step must implement all these types
        step: buildstep.BuildStep | GitStepMixin | CompositeStepMixin,
        ssh_private_key: IRenderable | str | None = None,
        ssh_host_key: IRenderable | str | None = None,
        ssh_known_hosts: IRenderable | str | None = None,
        git_credential_options: GitCredentialOptions | None = None,
    ) -> None:
        self.step = step

        super().__init__(ssh_private_key, ssh_host_key, ssh_known_hosts, git_credential_options)

    def _get_auth_data_path(self, data_workdir: str) -> str:
        # we can't use the workdir for temporary ssh-related files, because
        # it's needed when cloning repositories and git does not like the
        # destination directory being non-empty. We have to use separate
        # temporary directory for that data to ensure the confidentiality of it.
        # So instead of
        # '{path}/{to}/{workerbuilddir}/{workdir}/.buildbot-ssh-key'
        # we put the key in
        # '{path}/{to}/.{workerbuilddir}.{workdir}.buildbot/ssh-key'.

        # basename and dirname interpret the last element being empty for paths
        # ending with a slash
        assert (
            isinstance(self.step, buildstep.BuildStep)
            and self.step.build is not None
            and self.step.build.builder.config is not None
        )

        workerbuilddir = bytes2unicode(self.step.build.builder.config.workerbuilddir)
        workdir = data_workdir.rstrip('/\\')

        if self._path_module.isabs(workdir):
            parent_path = self._path_module.dirname(workdir)
        else:
            assert self.step.worker is not None
            parent_path = self._path_module.join(
                self.step.worker.worker_basedir, self._path_module.dirname(workdir)
            )

        basename = f'.{workerbuilddir}.{self._path_module.basename(workdir)}.buildbot'
        return self._path_module.join(parent_path, basename)

    def adjust_git_command_params_for_auth(
        self,
        full_command: list[str],
        full_env: dict[str, str],
        workdir: str,
        git_mixin: GitMixin,
    ) -> None:
        auth_data_path = self._get_auth_data_path(workdir)
        super().adjust_git_command_params_for_auth(
            full_command,
            full_env,
            workdir=auth_data_path,
            git_mixin=git_mixin,
        )

    @property
    def _path_module(self):
        assert isinstance(self.step, buildstep.BuildStep) and self.step.build is not None
        return self.step.build.path_module

    @property
    def _path_cls(self) -> type[PurePath]:
        assert isinstance(self.step, buildstep.BuildStep) and self.step.build is not None
        assert self.step.build.path_cls is not None
        return self.step.build.path_cls

    @property
    def _master(self):
        assert isinstance(self.step, buildstep.BuildStep) and self.step.master is not None
        return self.step.master

    async def _download_file(
        self,
        path: str,
        content: str,
        mode: int,
        workdir: str | None = None,
    ) -> None:
        assert isinstance(self.step, CompositeStepMixin)
        await self.step.downloadFileContentToWorker(
            path,
            content,
            mode=mode,
            workdir=workdir,
        )

    @async_to_deferred
    async def _dovccmd(
        self,
        command: list[str],
        initial_stdin: str | None = None,
        workdir: str | None = None,
    ) -> None:
        assert isinstance(self.step, GitStepMixin)
        await self.step._dovccmd(
            command=command,
            initialStdin=initial_stdin,
        )

    @async_to_deferred
    async def download_auth_files_if_needed(
        self,
        workdir: str,
        download_wrapper_script: bool = False,
    ) -> int:
        if self.ssh_private_key is None and self.git_credential_options is None:
            return RC_SUCCESS

        assert isinstance(self.step, CompositeStepMixin) and isinstance(self.step, GitMixin)

        workdir = self._get_auth_data_path(workdir)
        await self.step.runMkdir(workdir)

        return_code = await super().download_auth_files_if_needed(
            workdir=workdir,
            download_wrapper_script=(
                download_wrapper_script or not self.step.supportsSshPrivateKeyAsEnvOption
            ),
        )
        return return_code

    @async_to_deferred
    async def remove_auth_files_if_needed(self, workdir: str) -> int:
        if not self.did_download_auth_files:
            return RC_SUCCESS

        assert isinstance(self.step, CompositeStepMixin)
        await self.step.runRmdir(self._get_auth_data_path(workdir))
        return RC_SUCCESS


class GitServiceAuth(AbstractGitAuth):
    def __init__(
        self,
        service: GitPoller,
        ssh_private_key: IRenderable | str | None = None,
        ssh_host_key: IRenderable | str | None = None,
        ssh_known_hosts: IRenderable | str | None = None,
        git_credential_options: GitCredentialOptions | None = None,
    ) -> None:
        self._service = service

        super().__init__(ssh_private_key, ssh_host_key, ssh_known_hosts, git_credential_options)

    @property
    def _path_module(self):
        return os.path

    @property
    def _master(self):
        assert self._service.master is not None
        return self._service.master

    @async_to_deferred
    async def _dovccmd(
        self,
        command: list[str],
        initial_stdin: str | None = None,
        workdir: str | None = None,
    ) -> None:
        await self._service._dovccmd(
            command=command[0],
            args=command[1:],
            initial_stdin=initial_stdin,
            path=workdir,
            auth_files_path=workdir,  # this is ... not great
        )

    async def _download_file(
        self,
        path: str,
        content: str,
        mode: int,
        workdir: str | None = None,
    ) -> None:
        writeLocalFile(path, content, mode=mode)

    @async_to_deferred
    async def remove_auth_files_if_needed(self, workdir: str) -> int:
        if not self.did_download_auth_files:
            return RC_SUCCESS

        Path(self._get_ssh_private_key_path(workdir)).unlink(missing_ok=True)
        Path(self._get_ssh_host_key_path(workdir)).unlink(missing_ok=True)

        return RC_SUCCESS



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.process.results import SUCCESS
from buildbot.steps.source import gitlab
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectListdir
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import ExpectStat
from buildbot.test.util import config
from buildbot.test.util import sourcesteps


class TestGitLab(
    sourcesteps.SourceStepMixin, config.ConfigErrorsMixin, TestReactorMixin, unittest.TestCase
):
    stepClass = gitlab.GitLab

    def setUp(self):
        self.setup_test_reactor()
        self.sourceName = self.stepClass.__name__
        return self.setup_test_build_step()

    def setup_step(self, step, args, **kwargs):
        step = super().setup_step(step, args, **kwargs)
        step.build.properties.setProperty("source_branch", "ms-viewport", "gitlab source branch")
        step.build.properties.setProperty(
            "source_git_ssh_url",
            "git@gitlab.example.com:build/awesome_project.git",
            "gitlab source git ssh url",
        )
        step.build.properties.setProperty("source_project_id", 2337, "gitlab source project ID")
        step.build.properties.setProperty("target_branch", "master", "gitlab target branch")
        step.build.properties.setProperty(
            "target_git_ssh_url",
            "git@gitlab.example.com:mmusterman/awesome_project.git",
            "gitlab target git ssh url",
        )
        step.build.properties.setProperty("target_project_id", 239, "gitlab target project ID")
        return step

    def test_with_merge_branch(self):
        self.setup_step(
            self.stepClass(
                repourl='git@gitlab.example.com:mmusterman/awesome_project.git',
                mode='full',
                method='clean',
            ),
            {"branch": 'master', "revision": '12345678'},
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['git', '--version'])
            .stdout('git version 1.7.5')
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectListdir(dir='wkdir').files(['.git']).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'clean', '-f', '-f', '-d']).exit(0),
            # here we always ignore revision, and fetch the merge branch
            ExpectShell(
                workdir='wkdir',
                command=[
                    'git',
                    'fetch',
                    '-f',
                    '--progress',
                    'ssh://git@gitlab.example.com:22/build/awesome_project.git',
                    'ms-viewport',
                ],
            ).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'checkout', '-f', 'FETCH_HEAD']).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'checkout', '-B', 'ms-viewport']).exit(0),
            ExpectShell(workdir='wkdir', command=['git', 'rev-parse', 'HEAD'])
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', 'f6ad368298bd941e934a41f3babc827b2aa95a1d', 'GitLab')
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from sphinx.domains import Index
from sphinx.domains.std import StandardDomain


class PythonAPIIndex(Index):
    objecttype = 'class'
    name = 'apiindex'
    localname = 'Public API Index'
    shortname = 'classes'

    def generate(self, docnames=None):
        unsorted_objects = [
            (refname, entry.docname, entry.objtype)
            for (refname, entry) in self.domain.data['objects'].items()
            if entry.objtype in ['class', 'function']
        ]
        objects = sorted(unsorted_objects, key=lambda x: x[0].lower())

        entries = []

        for refname, docname, objtype in objects:
            if docnames and docname not in docnames:
                continue

            extra_info = objtype
            display_name = refname
            if objtype == 'function':
                display_name += '()'
            entries.append([display_name, 0, docname, refname, extra_info, '', ''])

        return [('', entries)], False


def setup(app):
    app.add_index_to_domain('py', PythonAPIIndex)
    StandardDomain.initial_data['labels']['apiindex'] = ('py-apiindex', '', 'Public API Index')
    StandardDomain.initial_data['anonlabels']['apiindex'] = ('py-apiindex', '')
    return {'parallel_read_safe': True, 'parallel_write_safe': True}



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import datetime
import json
import re
from typing import Any
from unittest import mock

import jwt
from autobahn.websocket.types import ConnectionDeny
from autobahn.websocket.types import ConnectionRequest
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import www
from buildbot.util import bytes2unicode
from buildbot.www import auth
from buildbot.www import ws


class WsResource(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = master = yield self.make_master(url="h:/a/b/", wantMq=True, wantGraphql=True)
        self.ws = ws.WsResource(master)
        self.proto = self.ws._factory.buildProtocol("me")
        self.proto.sendMessage = mock.Mock(spec=self.proto.sendMessage)

    def assert_called_with_json(self, obj, expected_json):
        jsonArg = obj.call_args[0][0]
        jsonArg = bytes2unicode(jsonArg)
        actual_json = json.loads(jsonArg)

        keys_to_pop = []
        for key in expected_json:
            if hasattr(expected_json[key], 'match'):
                keys_to_pop.append(key)
                regex = expected_json[key]
                value = actual_json[key]
                self.assertRegex(value, regex)

        for key in keys_to_pop:
            expected_json.pop(key)
            actual_json.pop(key)

        self.assertEqual(actual_json, expected_json)

    def test_ping(self):
        self.proto.onMessage(json.dumps({"cmd": 'ping', "_id": 1}), False)
        self.assert_called_with_json(self.proto.sendMessage, {"msg": "pong", "code": 200, "_id": 1})

    def test_bad_cmd(self):
        self.proto.onMessage(json.dumps({"cmd": 'poing', "_id": 1}), False)
        self.assert_called_with_json(
            self.proto.sendMessage,
            {"_id": 1, "code": 404, "error": "no such command type 'poing'"},
        )

    def test_no_cmd(self):
        self.proto.onMessage(json.dumps({"_id": 1}), False)
        self.assert_called_with_json(
            self.proto.sendMessage,
            {"_id": None, "code": 400, "error": "no 'cmd' in websocket frame"},
        )

    def test_too_many_arguments(self):
        self.proto.onMessage(json.dumps({"_id": 1, "cmd": 'ping', "foo": 'bar'}), False)
        self.assert_called_with_json(
            self.proto.sendMessage,
            {
                "_id": 1,
                "code": 400,
                "error": re.compile(".*Invalid method argument.*"),
            },
        )

    def test_no_id(self):
        self.proto.onMessage(json.dumps({"cmd": 'ping'}), False)
        self.assert_called_with_json(
            self.proto.sendMessage,
            {
                "_id": None,
                "code": 400,
                "error": "no '_id' or 'type' in websocket frame",
            },
        )

    def test_startConsuming(self):
        self.proto.onMessage(
            json.dumps({"cmd": 'startConsuming', "path": 'builds/*/*', "_id": 1}), False
        )
        self.assert_called_with_json(self.proto.sendMessage, {"msg": "OK", "code": 200, "_id": 1})
        self.master.mq.verifyMessages = False
        self.master.mq.callConsumer(("builds", "1", "new"), {"buildid": 1})
        self.assert_called_with_json(
            self.proto.sendMessage, {"k": "builds/1/new", "m": {"buildid": 1}}
        )

    def test_startConsumingBadPath(self):
        self.proto.onMessage(json.dumps({"cmd": 'startConsuming', "path": {}, "_id": 1}), False)
        self.assert_called_with_json(
            self.proto.sendMessage,
            {"_id": 1, "code": 400, "error": "invalid path format '{}'"},
        )

    def test_stopConsumingNotRegistered(self):
        self.proto.onMessage(
            json.dumps({"cmd": 'stopConsuming', "path": 'builds/*/*', "_id": 1}), False
        )
        self.assert_called_with_json(
            self.proto.sendMessage,
            {"_id": 1, "code": 400, "error": "path was not consumed 'builds/*/*'"},
        )

    def test_stopConsuming(self):
        self.proto.onMessage(
            json.dumps({"cmd": 'startConsuming', "path": 'builds/*/*', "_id": 1}), False
        )
        self.assert_called_with_json(self.proto.sendMessage, {"msg": "OK", "code": 200, "_id": 1})
        self.proto.onMessage(
            json.dumps({"cmd": 'stopConsuming', "path": 'builds/*/*', "_id": 2}), False
        )
        self.assert_called_with_json(self.proto.sendMessage, {"msg": "OK", "code": 200, "_id": 2})

    def build_token(self, expired: bool, user_info: dict[str, Any]) -> str:
        delta = datetime.timedelta(weeks=1)
        if expired:
            delta = -delta

        expiration = datetime.datetime.now(datetime.timezone.utc) + delta

        payload = {'user_info': user_info, 'exp': expiration}
        return jwt.encode(
            payload, self.master.www.site.session_secret, algorithm=auth.SESSION_SECRET_ALGORITHM
        )

    @defer.inlineCallbacks
    def test_on_connect_no_ssl(self):
        self.master.www = mock.Mock()
        self.master.www.authz = mock.Mock()
        self.master.www.authz.assertUserAllowed = mock.Mock(return_value=defer.succeed(None))

        request = ConnectionRequest(
            path='/ws',
            headers={'cookie': ''},
            peer='tcp:127.0.0.1:1234',
            host='localhost',
            origin='http://localhost',
            protocols=[],
            version=13,
            params={},
            extensions=[],
        )
        yield self.proto.onConnect(request)
        self.master.www.authz.assertUserAllowed.assert_called_once_with(
            'masters', 'get', {}, {'anonymous': True}
        )

    @defer.inlineCallbacks
    def test_on_connect_with_token(self):
        self.master.www = mock.Mock()
        self.master.www.site = mock.Mock()
        self.master.www.site.session_secret = 'secret'
        self.master.www.authz = mock.Mock()
        self.master.www.authz.assertUserAllowed = mock.Mock(return_value=defer.succeed(None))

        token = self.build_token(expired=False, user_info={'some': 'payload'})

        request = ConnectionRequest(
            path='/ws',
            headers={'cookie': f'TWISTED_SESSION={token}'},
            peer='tcp:127.0.0.1:1234',
            host='localhost',
            origin='http://localhost',
            protocols=[],
            version=13,
            params={},
            extensions=[],
        )

        yield self.proto.onConnect(request)
        self.master.www.authz.assertUserAllowed.assert_called_once_with(
            'masters', 'get', {}, {'some': 'payload'}
        )

    @defer.inlineCallbacks
    def test_on_connect_with_expired_token(self):
        self.master.www = mock.Mock()
        self.master.www.site = mock.Mock()
        self.master.www.site.session_secret = 'secret'
        self.master.www.authz = mock.Mock()
        self.master.www.authz.assertUserAllowed = mock.Mock(return_value=defer.succeed(None))

        token = self.build_token(expired=True, user_info={'some': 'payload'})

        request = ConnectionRequest(
            path='/ws',
            headers={'cookie': f'TWISTED_SESSION={token}'},
            peer='tcp:127.0.0.1:1234',
            host='localhost',
            origin='http://localhost',
            protocols=[],
            version=13,
            params={},
            extensions=[],
        )

        with self.assertRaises(ConnectionDeny) as cm:
            yield self.proto.onConnect(request)
        self.assertEqual(cm.exception.args, (403, 'Forbidden'))

    @defer.inlineCallbacks
    def test_on_connect_invalid_token(self):
        self.master.www = mock.Mock()
        self.master.www.site = mock.Mock()
        self.master.www.site.session_secret = 'secret'
        self.master.www.authz = mock.Mock()

        request = ConnectionRequest(
            path='/ws',
            headers={'cookie': 'TWISTED_SESSION=invalid_token'},
            peer='tcp:127.0.0.1:1234',
            host='localhost',
            origin='http://localhost',
            protocols=[],
            version=13,
            params={},
            extensions=[],
        )

        with self.assertRaises(ConnectionDeny) as cm:
            yield self.proto.onConnect(request)
        self.assertEqual(cm.exception.args, (403, 'Forbidden'))
        self.assertEqual(len(self.flushLoggedErrors(jwt.exceptions.DecodeError)), 1)

    @defer.inlineCallbacks
    def test_on_connect_with_ssl(self):
        self.master.www = mock.Mock()
        self.master.www.site = mock.Mock()
        self.master.www.site.session_secret = 'secret'
        self.master.www.authz = mock.Mock()
        self.master.www.authz.assertUserAllowed = mock.Mock(return_value=defer.succeed(None))

        self.proto.is_secure = mock.Mock(return_value=True)

        token = self.build_token(expired=False, user_info={'some': 'payload'})
        request = ConnectionRequest(
            path='/ws',
            headers={'cookie': f'TWISTED_SECURE_SESSION={token}'},
            peer='tcp:127.0.0.1:1234',
            host='localhost',
            origin='http://localhost',
            protocols=[],
            version=13,
            params={},
            extensions=[],
        )

        yield self.proto.onConnect(request)
        self.master.www.authz.assertUserAllowed.assert_called_once_with(
            'masters', 'get', {}, {'some': 'payload'}
        )

    @defer.inlineCallbacks
    def test_on_connect_different_path(self):
        self.master.www = mock.Mock()
        self.master.www.site = mock.Mock()
        self.master.www.site.session_secret = 'secret'
        self.master.www.authz = mock.Mock()
        self.master.www.authz.assertUserAllowed = mock.Mock(return_value=defer.succeed(None))

        token = self.build_token(expired=False, user_info={'some': 'payload'})

        request = ConnectionRequest(
            path='/custom/ws',
            headers={'cookie': f'TWISTED_SESSION_custom={token}'},
            peer='tcp:127.0.0.1:1234',
            host='localhost',
            origin='http://localhost',
            protocols=[],
            version=13,
            params={},
            extensions=[],
        )

        yield self.proto.onConnect(request)
        self.master.www.authz.assertUserAllowed.assert_called_once_with(
            'masters', 'get', {}, {'some': 'payload'}
        )

    @defer.inlineCallbacks
    def test_on_connect_direct_connection_deny(self):
        self.master.www = mock.Mock()
        self.master.www.site = mock.Mock()
        self.master.www.site.session_secret = 'secret'
        self.master.www.authz = mock.Mock()
        self.master.www.authz.assertUserAllowed = mock.Mock(
            side_effect=ConnectionDeny(403, "Forbidden")
        )

        request = ConnectionRequest(
            path='/ws',
            headers={'cookie': 'auth_token=valid'},
            peer='tcp:127.0.0.1:1234',
            host='localhost',
            origin='http://localhost',
            protocols=[],
            version=13,
            params={},
            extensions=[],
        )

        with self.assertRaises(ConnectionDeny) as cm:
            yield self.proto.onConnect(request)
        self.assertEqual(cm.exception.args, (403, 'Forbidden'))


class TestParseCookies(unittest.TestCase):
    def test_parse_cookies_single(self):
        result = ws.parse_cookies("name=value")
        self.assertEqual(result, {"name": "value"})

    def test_parse_cookies_multiple_comma(self):
        result = ws.parse_cookies("name1=value1,name2=value2")
        self.assertEqual(result, {"name1": "value1", "name2": "value2"})

    def test_parse_cookies_multiple_semicolon(self):
        result = ws.parse_cookies("name1=value1; name2=value2")
        self.assertEqual(result, {"name1": "value1", "name2": "value2"})

    def test_parse_cookies_mixed_separators(self):
        result = ws.parse_cookies("name1=value1,name2=value2; name3=value3")
        self.assertEqual(result, {"name1": "value1", "name2": "value2", "name3": "value3"})

    def test_parse_cookies_malformed(self):
        result = ws.parse_cookies("name1=value1; invalid; name2=value2")
        self.assertEqual(result, {"name1": "value1", "name2": "value2"})

    def test_parse_cookies_empty(self):
        result = ws.parse_cookies("")
        self.assertEqual(result, {})



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import os
import sys
from io import StringIO
from unittest import mock
from unittest.case import SkipTest

from twisted.internet import defer
from twisted.python import runtime
from twisted.trial import unittest

from buildbot.test.fake import fakemaster
from buildbot.test.fake import httpclientservice as fakehttpclientservice
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import config
from buildbot.util import kubeclientservice
from buildbot.util import service


class MockFileBase:
    file_mock_config: dict[str, str] = {}

    def setUp(self):
        self.patcher = mock.patch('buildbot.util.kubeclientservice.open', self.mock_open)
        self.patcher.start()

    def tearDown(self):
        self.patcher.stop()

    def mock_open(self, filename, mode=None, encoding='UTF-8'):
        filename_type = os.path.basename(filename)
        file_value = self.file_mock_config[filename_type]
        mock_open = mock.Mock(
            __enter__=mock.Mock(return_value=StringIO(file_value)), __exit__=mock.Mock()
        )
        return mock_open


class KubeClientServiceTestClusterConfig(MockFileBase, config.ConfigErrorsMixin, unittest.TestCase):
    file_mock_config = {'token': 'BASE64_TOKEN', 'namespace': 'buildbot_namespace'}

    def setUp(self):
        super().setUp()
        self.patch(kubeclientservice.os, 'environ', {'KUBERNETES_PORT': 'tcp://foo'})

    def patchExist(self, val):
        self.patch(kubeclientservice.os.path, 'exists', lambda x: val)

    def test_not_exists(self):
        self.patchExist(False)
        with self.assertRaisesConfigError('kube_dir not found:'):
            kubeclientservice.KubeInClusterConfigLoader()

    @defer.inlineCallbacks
    def test_basic(self):
        self.patchExist(True)
        config = kubeclientservice.KubeInClusterConfigLoader()
        yield config.startService()
        self.assertEqual(
            config.getConfig(),
            {
                'headers': {'Authorization': 'Bearer BASE64_TOKEN'},
                'master_url': 'https://foo',
                'namespace': 'buildbot_namespace',
                'verify': '/var/run/secrets/kubernetes.io/serviceaccount/ca.crt',
            },
        )


KUBE_CTL_PROXY_FAKE = """
import time
import sys

print("Starting to serve on 127.0.0.1:" + sys.argv[2])
sys.stdout.flush()
time.sleep(1000)
"""

KUBE_CTL_PROXY_FAKE_ERROR = """
import time
import sys

print("Issue with the config!", file=sys.stderr)
sys.stderr.flush()
sys.exit(1)
"""


class KubeClientServiceTestKubeHardcodedConfig(
    TestReactorMixin, config.ConfigErrorsMixin, unittest.TestCase
):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self, "http://localhost:8001"
        )
        yield self.master.startService()
        self.addCleanup(self.master.stopService)

    def test_basic(self):
        self.config = kubeclientservice.KubeHardcodedConfig(
            master_url="http://localhost:8001", namespace="default"
        )
        self.assertEqual(
            self.config.getConfig(),
            {'master_url': 'http://localhost:8001', 'namespace': 'default', 'headers': {}},
        )

    def test_cannot_pass_both_bearer_and_basic_auth(self):
        with self.assertRaises(RuntimeError):
            kubeclientservice.KubeHardcodedConfig(
                master_url="http://localhost:8001",
                namespace="default",
                verify="/path/to/pem",
                basicAuth="Bla",
                bearerToken="Bla",
            )


class KubeClientServiceTestKubeCtlProxyConfig(config.ConfigErrorsMixin, unittest.TestCase):
    def patchProxyCmd(self, cmd):
        if runtime.platformType != 'posix':
            self.config = None
            raise SkipTest('only posix platform is supported by this test')
        self.patch(
            kubeclientservice.KubeCtlProxyConfigLoader,
            'kube_ctl_proxy_cmd',
            [sys.executable, "-c", cmd],
        )

    def tearDown(self):
        if self.config is not None and self.config.running:
            return self.config.stopService()
        return None

    @defer.inlineCallbacks
    def test_basic(self):
        self.patchProxyCmd(KUBE_CTL_PROXY_FAKE)
        self.config = kubeclientservice.KubeCtlProxyConfigLoader()
        yield self.config.startService()
        self.assertEqual(
            self.config.getConfig(), {'master_url': 'http://localhost:8001', 'namespace': 'default'}
        )

    @defer.inlineCallbacks
    def test_config_args(self):
        self.patchProxyCmd(KUBE_CTL_PROXY_FAKE)
        self.config = kubeclientservice.KubeCtlProxyConfigLoader(
            proxy_port=8002, namespace="system"
        )
        yield self.config.startService()
        self.assertEqual(self.config.kube_proxy_output, b'Starting to serve on 127.0.0.1:8002')
        self.assertEqual(
            self.config.getConfig(), {'master_url': 'http://localhost:8002', 'namespace': 'system'}
        )
        yield self.config.stopService()

    @defer.inlineCallbacks
    def test_reconfig(self):
        self.patchProxyCmd(KUBE_CTL_PROXY_FAKE)
        self.config = kubeclientservice.KubeCtlProxyConfigLoader(
            proxy_port=8002, namespace="system"
        )
        yield self.config.startService()
        self.assertEqual(self.config.kube_proxy_output, b'Starting to serve on 127.0.0.1:8002')
        self.assertEqual(
            self.config.getConfig(), {'master_url': 'http://localhost:8002', 'namespace': 'system'}
        )
        yield self.config.reconfigService(proxy_port=8003, namespace="system2")
        self.assertEqual(self.config.kube_proxy_output, b'Starting to serve on 127.0.0.1:8003')
        self.assertEqual(
            self.config.getConfig(), {'master_url': 'http://localhost:8003', 'namespace': 'system2'}
        )
        yield self.config.stopService()

    @defer.inlineCallbacks
    def test_config_with_error(self):
        self.patchProxyCmd(KUBE_CTL_PROXY_FAKE_ERROR)
        self.config = kubeclientservice.KubeCtlProxyConfigLoader()
        with self.assertRaises(RuntimeError):
            yield self.config.startService()


class KubeClientServiceTest(unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.parent = service.BuildbotService(name="parent")

        @defer.inlineCallbacks
        def cleanup():
            if self.parent.running:
                yield self.parent.stopService()

        self.addCleanup(cleanup)

        self.client = kubeclientservice.KubeClientService()
        yield self.client.setServiceParent(self.parent)

    @defer.inlineCallbacks
    def test_stopped(self):
        worker = mock.Mock(name="worker1")
        config = service.BuildbotService(name="config")

        yield self.client.register(worker, config)
        self.assertEqual(config.running, 0)
        yield self.client.unregister(worker)
        self.assertEqual(config.running, 0)

    @defer.inlineCallbacks
    def test_started(self):
        yield self.parent.startService()

        worker = mock.Mock(name="worker1")
        config = service.BuildbotService(name="config")

        yield self.client.register(worker, config)
        self.assertEqual(config.running, 1)
        yield self.client.unregister(worker)
        self.assertEqual(config.running, 0)

    @defer.inlineCallbacks
    def test_started_but_stop(self):
        yield self.parent.startService()

        worker = mock.Mock(name="worker1")
        config = service.BuildbotService(name="config")

        yield self.client.register(worker, config)
        self.assertEqual(config.running, 1)

        yield self.parent.stopService()
        self.assertEqual(config.running, 0)

    @defer.inlineCallbacks
    def test_stopped_but_start(self):
        worker = mock.Mock(name="worker1")
        config = service.BuildbotService(name="config")

        yield self.client.register(worker, config)
        self.assertEqual(config.running, 0)

        yield self.parent.startService()
        self.assertEqual(config.running, 1)

        yield self.parent.stopService()
        self.assertEqual(config.running, 0)

    @defer.inlineCallbacks
    def test_two_workers(self):
        yield self.parent.startService()

        worker1 = mock.Mock(name="worker1")
        worker2 = mock.Mock(name="worker2")
        config = service.BuildbotService(name="config")

        yield self.client.register(worker1, config)
        self.assertEqual(config.running, 1)
        yield self.client.register(worker2, config)
        self.assertEqual(config.running, 1)
        yield self.client.unregister(worker1)
        self.assertEqual(config.running, 1)
        yield self.client.unregister(worker2)
        self.assertEqual(config.running, 0)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.schedulers.canceller import OldBuildCanceller
from buildbot.schedulers.canceller import _OldBuildFilterSet
from buildbot.schedulers.canceller import _OldBuildrequestTracker
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.util.ssfilter import SourceStampFilter


class TestFilterSet(unittest.TestCase):
    def test_empty_filter(self):
        filter = _OldBuildFilterSet()
        self.assertFalse(filter.is_matched('builder', {'prop': 'value'}))

    @parameterized.expand([
        ('other_builder', 'builder2', {'project': 'p', 'repository': 'r'}, False),
        ('nothing', 'builder1', {'project': 'value_other', 'repository': 'value_other'}, False),
        ('single1', 'builder1', {'project': 'p', 'repository': 'value_other'}, True),
        ('single2', 'builder1', {'project': 'value_other', 'repository': 'r'}, True),
        ('all', 'builder1', {'project': 'p', 'repository': 'r'}, True),
    ])
    def test_multiple_filters_on_builder(self, name, builder, props, expected):
        filter = _OldBuildFilterSet()
        filter.add_filter(['builder1'], SourceStampFilter(project_eq='p'))
        filter.add_filter(['builder1'], SourceStampFilter(repository_eq='r'))

        self.assertEqual(filter.is_matched(builder, props), expected)


class TestOldBuildrequestTracker(TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        filter = _OldBuildFilterSet()

        ss_filter = SourceStampFilter(
            codebase_eq=['cb1', 'cb2'], repository_eq=['rp1', 'rp2'], branch_eq=['br1', 'br2']
        )
        filter.add_filter(['bldr1', 'bldr2'], ss_filter)
        self.cancellations = []
        self.tracker = _OldBuildrequestTracker(
            self.reactor, filter, lambda ss: ss['branch'], self.on_cancel
        )

    def on_cancel(self, brid):
        self.cancellations.append(brid)

    def assert_cancelled(self, cancellations):
        self.assertEqual(self.cancellations, cancellations)
        self.cancellations = []

    def create_ss_dict(self, project, codebase, repository, branch):
        # Changes have the same structure for the attributes that we're using, so we reuse this
        # function for changes.
        return {
            'project': project,
            'codebase': codebase,
            'repository': repository,
            'branch': branch,
        }

    def test_unknown_branch_not_tracked(self):
        ss_dicts = [self.create_ss_dict('pr1', 'cb1', 'rp1', None)]

        self.tracker.on_new_buildrequest(10, 'bldr1', ss_dicts)
        self.assertFalse(self.tracker.is_buildrequest_tracked(10))

    def test_multi_codebase_unknown_branch_not_tracked(self):
        ss_dicts = [
            self.create_ss_dict('pr1', 'cb1', 'rp1', None),
            self.create_ss_dict('pr2', 'cb2', 'rp2', 'br2'),
        ]

        self.tracker.on_new_buildrequest(10, 'bldr1', ss_dicts)
        self.assertFalse(self.tracker.is_buildrequest_tracked(10))

    def test_unmatched_ss_not_tracked(self):
        ss_dicts = [self.create_ss_dict('pr1', 'cb1', 'rp1', 'untracked')]

        self.tracker.on_new_buildrequest(10, 'bldr1', ss_dicts)
        self.assertFalse(self.tracker.is_buildrequest_tracked(10))

    def test_multi_codebase_unmatched_ss_not_tracked(self):
        ss_dicts = [
            self.create_ss_dict('pr1', 'cb1', 'rp1', 'untracked'),
            self.create_ss_dict('pr2', 'cb2', 'rp2', 'untracked'),
        ]

        self.tracker.on_new_buildrequest(10, 'bldr1', ss_dicts)
        self.assertFalse(self.tracker.is_buildrequest_tracked(10))

    def test_multi_codebase_tracks_if_at_least_one_ss_match(self):
        ss_dicts = [
            self.create_ss_dict('pr1', 'cb1', 'rp1', 'untracked'),
            self.create_ss_dict('pr2', 'cb2', 'rp2', 'br2'),
        ]

        self.tracker.on_new_buildrequest(10, 'bldr1', ss_dicts)
        self.assertTrue(self.tracker.is_buildrequest_tracked(10))

    def test_cancel_buildrequest(self):
        ss_dict = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br1')
        not_matching_ss_dict = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br2')

        self.tracker.on_new_buildrequest(1, 'bldr1', [ss_dict])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

        self.reactor.advance(1)

        self.tracker.on_change(not_matching_ss_dict)
        self.assert_cancelled([])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

        self.reactor.advance(1)

        self.tracker.on_change(ss_dict)
        self.assert_cancelled([])
        self.tracker.on_new_buildrequest(2, 'bldr1', [ss_dict])
        self.assert_cancelled([1])
        self.assertFalse(self.tracker.is_buildrequest_tracked(1))
        self.tracker.on_complete_buildrequest(1)

        self.reactor.advance(1)

        self.tracker.on_change(ss_dict)
        self.tracker.on_new_buildrequest(3, 'bldr1', [ss_dict])
        self.assert_cancelled([2])
        self.tracker.on_complete_buildrequest(2)

    def test_cancel_buildrequest_identical_times(self):
        ss_dict = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br1')
        not_matching_ss_dict = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br2')

        self.tracker.on_new_buildrequest(1, 'bldr1', [ss_dict])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

        self.tracker.on_change(not_matching_ss_dict)
        self.assert_cancelled([])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

        self.tracker.on_change(ss_dict)
        self.tracker.on_new_buildrequest(2, 'bldr1', [ss_dict])
        self.assert_cancelled([])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

        self.tracker.on_change(ss_dict)
        self.tracker.on_new_buildrequest(3, 'bldr1', [ss_dict])
        self.assert_cancelled([])

    def test_not_cancel_finished_buildrequest(self):
        ss_dict = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br1')

        self.tracker.on_new_buildrequest(1, 'bldr1', [ss_dict])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

        self.reactor.advance(1)

        self.tracker.on_complete_buildrequest(1)
        self.assertFalse(self.tracker.is_buildrequest_tracked(1))

        self.reactor.advance(1)

        self.tracker.on_change(ss_dict)
        self.tracker.on_new_buildrequest(2, 'bldr1', [ss_dict])
        self.assert_cancelled([])
        self.assertFalse(self.tracker.is_buildrequest_tracked(1))

    def test_not_cancel_buildrequest_too_new(self):
        ss_dict = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br1')

        self.tracker.on_change(ss_dict)
        self.tracker.on_new_buildrequest(1, 'bldr1', [ss_dict])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

        self.reactor.advance(1)

        self.tracker.on_new_buildrequest(2, 'bldr1', [ss_dict])
        self.assert_cancelled([])
        self.assertTrue(self.tracker.is_buildrequest_tracked(2))

        self.reactor.advance(1)

        self.tracker.on_complete_buildrequest(1)
        self.tracker.on_complete_buildrequest(2)

    def test_not_cancel_buildrequest_different_builder(self):
        ss_dict = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br1')

        self.tracker.on_change(ss_dict)
        self.tracker.on_new_buildrequest(1, 'bldr1', [ss_dict])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

        self.reactor.advance(1)

        self.tracker.on_change(ss_dict)
        self.tracker.on_new_buildrequest(2, 'bldr2', [ss_dict])
        self.assert_cancelled([])
        self.assertTrue(self.tracker.is_buildrequest_tracked(2))

        self.reactor.advance(1)

        self.tracker.on_complete_buildrequest(1)
        self.tracker.on_complete_buildrequest(2)

    @parameterized.expand([
        ('first', True),
        ('second', False),
    ])
    def test_cancel_multi_codebase_buildrequest(self, name, cancel_first_ss):
        ss_dict1 = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br1')
        ss_dict2 = self.create_ss_dict('pr2', 'cb2', 'rp2', 'br2')
        not_matching_ss_dict = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br2')

        self.tracker.on_new_buildrequest(1, 'bldr1', [ss_dict1, ss_dict2])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

        self.reactor.advance(1)

        self.tracker.on_change(not_matching_ss_dict)
        self.tracker.on_new_buildrequest(2, 'bldr1', [not_matching_ss_dict])
        self.assert_cancelled([])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

        self.reactor.advance(1)

        self.tracker.on_change(ss_dict1 if cancel_first_ss else ss_dict2)
        self.tracker.on_new_buildrequest(3, 'bldr1', [ss_dict1 if cancel_first_ss else ss_dict2])
        self.assert_cancelled([1])
        self.assertFalse(self.tracker.is_buildrequest_tracked(1))
        self.tracker.on_complete_buildrequest(1)

        self.reactor.advance(1)

        self.tracker.on_change(ss_dict1)
        self.tracker.on_new_buildrequest(4, 'bldr1', [ss_dict1])
        self.tracker.on_change(ss_dict2)
        self.tracker.on_new_buildrequest(5, 'bldr1', [ss_dict2])
        self.assert_cancelled([3])
        self.tracker.on_complete_buildrequest(3)

    def test_cancel_multi_codebase_buildrequest_ignores_non_matching_change_in_tracked_br(self):
        ss_dict1 = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br1')
        non_matched_ss_dict = self.create_ss_dict('pr1', 'cb1', 'rp1', 'brZ')

        self.tracker.on_new_buildrequest(1, 'bldr1', [ss_dict1, non_matched_ss_dict])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

        self.reactor.advance(1)

        self.tracker.on_change(non_matched_ss_dict)
        self.tracker.on_new_buildrequest(2, 'bldr1', [non_matched_ss_dict])
        self.assert_cancelled([])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))

    def test_cancel_multiple_buildrequests(self):
        ss_dict = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br1')
        not_matching_ss_dict = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br2')

        self.tracker.on_new_buildrequest(1, 'bldr1', [ss_dict])
        self.tracker.on_new_buildrequest(2, 'bldr1', [ss_dict])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))
        self.assertTrue(self.tracker.is_buildrequest_tracked(2))

        self.reactor.advance(1)

        self.tracker.on_change(not_matching_ss_dict)
        self.tracker.on_new_buildrequest(3, 'bldr1', [not_matching_ss_dict])
        self.assert_cancelled([])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))
        self.assertTrue(self.tracker.is_buildrequest_tracked(2))

        self.reactor.advance(1)

        self.tracker.on_change(ss_dict)
        self.tracker.on_new_buildrequest(4, 'bldr1', [ss_dict])
        self.assert_cancelled([1, 2])
        self.assertFalse(self.tracker.is_buildrequest_tracked(1))
        self.assertFalse(self.tracker.is_buildrequest_tracked(2))
        self.tracker.on_complete_buildrequest(1)
        self.tracker.on_complete_buildrequest(2)

        self.reactor.advance(1)

        self.tracker.on_change(ss_dict)
        self.tracker.on_new_buildrequest(5, 'bldr1', [ss_dict])
        self.assert_cancelled([4])
        self.tracker.on_complete_buildrequest(4)

    def test_cancel_multi_codebase_multiple_buildrequests(self):
        ss_dict1 = self.create_ss_dict('pr1', 'cb1', 'rp1', 'br1')
        ss_dict2 = self.create_ss_dict('pr2', 'cb2', 'rp2', 'br2')
        ss_dict3 = self.create_ss_dict('pr3', 'cb3', 'rp3', 'br3')

        self.tracker.on_new_buildrequest(1, 'bldr1', [ss_dict1, ss_dict2])
        self.tracker.on_new_buildrequest(2, 'bldr1', [ss_dict1, ss_dict3])
        self.tracker.on_new_buildrequest(3, 'bldr1', [ss_dict2, ss_dict3])
        self.assertTrue(self.tracker.is_buildrequest_tracked(1))
        self.assertTrue(self.tracker.is_buildrequest_tracked(2))
        self.assertTrue(self.tracker.is_buildrequest_tracked(3))
        self.assert_cancelled([])

        self.reactor.advance(1)

        self.tracker.on_change(ss_dict1)
        self.tracker.on_new_buildrequest(4, 'bldr1', [ss_dict1])
        self.assert_cancelled([1, 2])
        self.assertFalse(self.tracker.is_buildrequest_tracked(1))
        self.assertFalse(self.tracker.is_buildrequest_tracked(2))
        self.assertTrue(self.tracker.is_buildrequest_tracked(3))
        self.tracker.on_complete_buildrequest(1)
        self.tracker.on_complete_buildrequest(2)

        self.reactor.advance(1)

        self.tracker.on_change(ss_dict1)
        self.tracker.on_new_buildrequest(5, 'bldr1', [ss_dict1])
        self.assert_cancelled([4])
        self.tracker.on_complete_buildrequest(4)


class TestOldBuildCancellerUtils(ConfigErrorsMixin, unittest.TestCase):
    @parameterized.expand([
        ('only_builder', [(['bldr'], SourceStampFilter())]),
        ('with_codebase', [(['bldr'], SourceStampFilter(codebase_eq=['value']))]),
        ('with_repository', [(['bldr'], SourceStampFilter(repository_eq=['value']))]),
        ('with_branch', [(['bldr'], SourceStampFilter(branch_eq=['value']))]),
        (
            'all',
            [
                (
                    ['bldr'],
                    SourceStampFilter(
                        codebase_eq=['v1', 'v2'], repository_eq=['v1', 'v2'], branch_eq=['v1', 'v2']
                    ),
                )
            ],
        ),
    ])
    def test_check_filters_valid(self, name, filters):
        OldBuildCanceller.check_filters(filters)

    @parameterized.expand([
        ('dict', {}),
        ('list_list', [[]]),
    ])
    def test_check_filters_not_dict(self, name, value):
        with self.assertRaisesConfigError('The filters argument must be a list of tuples'):
            OldBuildCanceller.check_filters(value)

    def test_check_filters_invalid_uple(self):
        with self.assertRaisesConfigError('must be a list of tuples each of which'):
            OldBuildCanceller.check_filters([('a', 'b', 'c')])
        with self.assertRaisesConfigError('must be a list of tuples each of which'):
            OldBuildCanceller.check_filters([('a',)])

    @parameterized.expand([
        ('dict', {}, 'filter builders must be list of strings or a string'),
        ('list_int', [1], 'Value of filter builders must be string'),
    ])
    def test_check_builders_keys_not_list(self, name, value, error):
        with self.assertRaisesConfigError(error):
            OldBuildCanceller.check_filters([(value, SourceStampFilter())])


class TestOldBuildCanceller(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantMq=True, wantData=True, wantDb=True)
        self.master.mq.verifyMessages = False

        yield self.insert_test_data()
        self._cancelled_build_ids = []

        yield self.master.startService()
        self.addCleanup(self.master.stopService)

    def create_ss_dict(self, project, codebase, repository, branch):
        # Changes have the same structure for the attributes that we're using, so we reuse this
        # function for changes.
        return {
            'project': project,
            'codebase': codebase,
            'repository': repository,
            'branch': branch,
        }

    @defer.inlineCallbacks
    def insert_test_data(self):
        yield self.master.db.insert_test_data([
            fakedb.Master(id=92),
            fakedb.Worker(id=13, name='wrk'),
            fakedb.Builder(id=79, name='builder1'),
            fakedb.Builder(id=80, name='builder2'),
            fakedb.Builder(id=81, name='builder3'),
            fakedb.Buildset(id=98, results=None, reason="reason98"),
            fakedb.BuildsetSourceStamp(buildsetid=98, sourcestampid=234),
            fakedb.SourceStamp(
                id=234,
                revision='revision1',
                project='project1',
                codebase='codebase1',
                repository='repository1',
                branch='branch1',
            ),
            fakedb.BuildRequest(id=10, buildsetid=98, builderid=79),
            fakedb.Build(
                id=19,
                number=1,
                builderid=79,
                buildrequestid=10,
                workerid=13,
                masterid=92,
                results=None,
                state_string="state1",
            ),
            fakedb.Buildset(id=99, results=None, reason="reason99"),
            fakedb.BuildsetSourceStamp(buildsetid=99, sourcestampid=235),
            fakedb.SourceStamp(
                id=235,
                revision='revision2',
                project='project2',
                codebase='codebase2',
                repository='repository2',
                branch='branch2',
            ),
            fakedb.BuildRequest(id=11, buildsetid=99, builderid=80),
            fakedb.Build(
                id=20,
                number=1,
                builderid=80,
                buildrequestid=11,
                workerid=13,
                masterid=92,
                results=None,
                state_string="state2",
            ),
            fakedb.Buildset(id=100, results=None, reason="reason100"),
            fakedb.BuildsetSourceStamp(buildsetid=100, sourcestampid=236),
            fakedb.SourceStamp(
                id=236,
                revision='revision2',
                project='project2',
                codebase='codebase2',
                repository='repository2',
                branch='refs/changes/10/12310/2',
            ),
            fakedb.BuildRequest(id=12, buildsetid=100, builderid=81),
            fakedb.Build(
                id=21,
                number=1,
                builderid=81,
                buildrequestid=12,
                workerid=13,
                masterid=92,
                results=None,
                state_string="state3",
            ),
        ])

    @defer.inlineCallbacks
    def setup_canceller_with_filters(self):
        self.canceller = OldBuildCanceller(
            'canceller',
            [
                (['builder1'], SourceStampFilter(branch_eq=['branch1'])),
                (['builder2'], SourceStampFilter(branch_eq=['branch2'])),
                (['builder3'], SourceStampFilter()),
            ],
        )
        yield self.canceller.setServiceParent(self.master)

    @defer.inlineCallbacks
    def setup_canceller_with_no_filters(self):
        self.canceller = OldBuildCanceller('canceller', [])
        yield self.canceller.setServiceParent(self.master)

    def assert_cancelled(self, cancellations):
        expected_productions = []
        for id in cancellations:
            expected_productions.append((
                ('control', 'buildrequests', str(id), 'cancel'),
                {'reason': 'Build request has been obsoleted by a newer commit'},
            ))

        self.master.mq.assertProductions(expected_productions)

    @defer.inlineCallbacks
    def test_buildrequest_no_branch(self):
        yield self.setup_canceller_with_filters()

        self.reactor.advance(1)

        ss_dict = self.create_ss_dict('project1', 'codebase1', 'repository1', None)

        yield self.master.db.insert_test_data([
            fakedb.Buildset(id=199, results=None, reason='reason99'),
            fakedb.BuildsetSourceStamp(buildsetid=199, sourcestampid=240),
            fakedb.SourceStamp(
                id=240,
                revision='revision1',
                project='project1',
                codebase='codebase1',
                repository='repository1',
                branch=None,
            ),
            fakedb.BuildRequest(id=14, buildsetid=199, builderid=79),
        ])

        self.master.mq.callConsumer(('changes', '123', 'new'), ss_dict)
        self.master.mq.callConsumer(
            ('buildrequests', '14', 'new'),
            {'buildrequestid': 14, 'builderid': 79, 'buildsetid': 199},
        )
        self.assert_cancelled([])

    @defer.inlineCallbacks
    def test_cancel_buildrequest_after_new_commit_with_buildrequest(self):
        yield self.setup_canceller_with_filters()

        self.reactor.advance(1)

        ss_dict = self.create_ss_dict('project1', 'codebase1', 'repository1', 'branch1')

        yield self.master.db.insert_test_data([
            fakedb.Buildset(id=199, results=None, reason='reason99'),
            fakedb.BuildsetSourceStamp(buildsetid=199, sourcestampid=240),
            fakedb.SourceStamp(
                id=240,
                revision='revision240',
                project='project1',
                codebase='codebase1',
                repository='repository1',
                branch='branch1',
            ),
            fakedb.BuildRequest(id=14, buildsetid=199, builderid=79),
        ])

        self.master.mq.callConsumer(('changes', '123', 'new'), ss_dict)
        self.master.mq.callConsumer(
            ('buildrequests', '14', 'new'),
            {'buildrequestid': 14, 'builderid': 79, 'buildsetid': 199},
        )
        self.assert_cancelled([10])

        self.reactor.advance(1)

        yield self.master.db.insert_test_data([
            fakedb.Buildset(id=200, results=None, reason='reason100'),
            fakedb.BuildsetSourceStamp(buildsetid=200, sourcestampid=241),
            fakedb.SourceStamp(
                id=241,
                revision='revision241',
                project='project1',
                codebase='codebase1',
                repository='repository1',
                branch='branch1',
            ),
            fakedb.BuildRequest(id=15, buildsetid=200, builderid=79),
        ])

        self.master.mq.callConsumer(('changes', '124', 'new'), ss_dict)
        self.master.mq.callConsumer(
            ('buildrequests', '15', 'new'),
            {'buildrequestid': 15, 'builderid': 79, 'buildsetid': 200},
        )
        self.assert_cancelled([14])

    @defer.inlineCallbacks
    def test_no_cancel_buildrequest_after_only_new_commit(self):
        yield self.setup_canceller_with_filters()

        self.reactor.advance(1)

        ss_dict = self.create_ss_dict('project1', 'codebase1', 'repository1', 'branch1')

        self.master.mq.callConsumer(('changes', '123', 'new'), ss_dict)
        self.assert_cancelled([])

    @defer.inlineCallbacks
    def test_cancel_buildrequest_after_new_commit_gerrit_branch_filter(self):
        yield self.setup_canceller_with_filters()

        self.reactor.advance(1)

        ss_dict = self.create_ss_dict(
            'project2', 'codebase2', 'repository2', 'refs/changes/10/12310/3'
        )
        yield self.master.db.insert_test_data([
            fakedb.Buildset(id=199, results=None, reason='reason99'),
            fakedb.BuildsetSourceStamp(buildsetid=199, sourcestampid=240),
            fakedb.SourceStamp(
                id=240,
                revision='revision240',
                project='project2',
                codebase='codebase2',
                repository='repository2',
                branch='refs/changes/10/12310/3',
            ),
            fakedb.BuildRequest(id=14, buildsetid=199, builderid=81),
        ])

        self.master.mq.callConsumer(('changes', '123', 'new'), ss_dict)
        self.master.mq.callConsumer(
            ('buildrequests', '14', 'new'),
            {'buildrequestid': 14, 'builderid': 81, 'buildsetid': 199},
        )
        self.assert_cancelled([12])

        self.reactor.advance(1)

        yield self.master.db.insert_test_data([
            fakedb.Buildset(id=200, results=None, reason='reason100'),
            fakedb.BuildsetSourceStamp(buildsetid=200, sourcestampid=241),
            fakedb.SourceStamp(
                id=241,
                revision='revision241',
                project='project2',
                codebase='codebase2',
                repository='repository2',
                branch='refs/changes/10/12310/3',
            ),
            fakedb.BuildRequest(id=15, buildsetid=200, builderid=81),
        ])
        self.master.mq.callConsumer(('changes', '124', 'new'), ss_dict)
        self.master.mq.callConsumer(
            ('buildrequests', '15', 'new'),
            {'buildrequestid': 15, 'builderid': 81, 'buildsetid': 200},
        )
        self.assert_cancelled([14])

    @defer.inlineCallbacks
    def test_build_finished_then_new_commit_no_cancel(self):
        yield self.setup_canceller_with_filters()

        self.reactor.advance(1)

        ss_dict = self.create_ss_dict('project1', 'codebase1', 'repository1', 'branch1')

        self.master.mq.callConsumer(('buildrequests', '10', 'complete'), {'buildrequestid': 10})
        self.master.mq.callConsumer(('changes', '123', 'new'), ss_dict)
        self.assert_cancelled([])

    @defer.inlineCallbacks
    def test_reconfig_no_longer_matched_tracked_build_cancelled(self):
        yield self.setup_canceller_with_filters()

        self.reactor.advance(1)

        ss_dict = self.create_ss_dict('project1', 'codebase1', 'repository1', 'branch1')

        yield self.canceller.reconfigService('canceller', [])

        yield self.master.db.insert_test_data([
            fakedb.Buildset(id=199, results=None, reason='reason99'),
            fakedb.BuildsetSourceStamp(buildsetid=199, sourcestampid=242),
            fakedb.SourceStamp(
                id=242,
                revision='revision242',
                project='project1',
                codebase='codebase1',
                repository='repository1',
                branch='branch1',
            ),
            fakedb.BuildRequest(id=14, buildsetid=199, builderid=79),
        ])

        self.master.mq.callConsumer(('changes', '123', 'new'), ss_dict)
        self.master.mq.callConsumer(
            ('buildrequests', '14', 'new'),
            {'buildrequestid': 14, 'builderid': 79, 'buildsetid': 199},
        )
        self.assert_cancelled([10])

        self.reactor.advance(1)

        yield self.master.db.insert_test_data([
            fakedb.Buildset(id=200, results=None, reason='reason100'),
            fakedb.BuildsetSourceStamp(buildsetid=200, sourcestampid=240),
            fakedb.SourceStamp(
                id=240,
                revision='revision240',
                project='project1',
                codebase='codebase1',
                repository='repository1',
                branch='branch1',
            ),
            fakedb.BuildRequest(id=15, buildsetid=200, builderid=79),
        ])
        self.master.mq.callConsumer(('changes', '124', 'new'), ss_dict)
        self.master.mq.callConsumer(
            ('buildrequests', '15', 'new'),
            {'buildrequestid': 15, 'builderid': 79, 'buildsetid': 200},
        )
        self.assert_cancelled([])

    @defer.inlineCallbacks
    def test_reconfig_defers_finished_builds_to_after_registration(self):
        # We need to make sure that during reconfiguration any finished build messages are not
        # acted before the build is tracked

        yield self.setup_canceller_with_no_filters()

        # Setup controllable blocking wait on canceller._on_buildrequest_new
        on_buildrequest_new_d = defer.Deferred()
        on_buildrequest_new_original = self.canceller._on_buildrequest_new
        on_buildrequest_new_breq_ids = []

        @defer.inlineCallbacks
        def waiting_on_buildrequest_new(key, breq):
            on_buildrequest_new_breq_ids.append(breq['buildrequestid'])
            if not on_buildrequest_new_d.called:
                yield on_buildrequest_new_d
            yield on_buildrequest_new_original(key, breq)

        self.canceller._on_buildrequest_new = waiting_on_buildrequest_new

        # Start reconfig. We verify that we actually blocked in on_buildrequest_new
        d = self.canceller.reconfigService(
            'canceller',
            [
                {'builders': ['builder1'], 'branch_eq': ['branch1']},
                {'builders': ['builder2'], 'branch_eq': ['branch2']},
            ],
        )

        self.assertEqual(on_buildrequest_new_breq_ids, [10])
        self.assertFalse(d.called)

        # The buildrequest complete messages should be queued
        self.master.mq.callConsumer(('buildrequests', '10', 'complete'), {'buildrequestid': 10})
        self.master.mq.callConsumer(('buildrequests', '11', 'complete'), {'buildrequestid': 11})

        # Unblock reconfigService
        on_buildrequest_new_d.callback(None)
        yield d
        self.assertEqual(on_buildrequest_new_breq_ids, [10, 11, 12])

        self.assertFalse(self.canceller._build_tracker.is_buildrequest_tracked(10))
        self.assertFalse(self.canceller._build_tracker.is_buildrequest_tracked(11))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.test.fakedb.row import Row


class Codebase(Row):
    table = "codebases"

    id_column = 'id'
    hashedColumns = [('name_hash', ('name',))]

    def __init__(
        self,
        *,
        id: int | None = None,
        projectid: int | None = None,
        name: str = 'fake_codebase',
        name_hash: str | None = None,
        slug: str | None = None,
    ) -> None:
        if slug is None:
            slug = name
        super().__init__(
            id=id,
            projectid=projectid,
            name=name,
            name_hash=name_hash,
            slug=slug,
        )


class CodebaseCommit(Row):
    table = "codebase_commits"

    id_column = 'id'

    def __init__(
        self,
        *,
        id: int,
        codebaseid: int,
        author: str = 'author1',
        committer: str | None = 'committer1',
        comments: str = 'comments1',
        when_timestamp: int = 1234567,
        revision: str | None = None,
        parent_commitid: int | None = None,
    ):
        if revision is None:
            revision = f'rev{id}'
        super().__init__(
            id=id,
            codebaseid=codebaseid,
            author=author,
            committer=committer,
            comments=comments,
            when_timestamp=when_timestamp,
            revision=revision,
            parent_commitid=parent_commitid,
        )


class CodebaseBranch(Row):
    table = "codebase_branches"

    id_column = 'id'
    hashedColumns = [('name_hash', ('name',))]

    def __init__(
        self,
        *,
        id: int | None = None,
        codebaseid: int | None = None,
        name: str = 'fake_branch',
        name_hash: str | None = None,
        commitid: int | None = None,
        last_timestamp: int = 1234567,
    ):
        super().__init__(
            id=id,
            codebaseid=codebaseid,
            name=name,
            name_hash=name_hash,
            commitid=commitid,
            last_timestamp=last_timestamp,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial.unittest import SkipTest

from buildbot.config import BuilderConfig
from buildbot.process.factory import BuildFactory
from buildbot.schedulers.forcesched import ForceScheduler
from buildbot.steps.download_secret_to_worker import DownloadSecretsToWorker
from buildbot.steps.download_secret_to_worker import RemoveWorkerFileSecret
from buildbot.test.util.integration import RunMasterBase


class DownloadSecretsBase(RunMasterBase):
    def setUp(self):
        self.temp_dir = os.path.abspath(self.mktemp())
        os.mkdir(self.temp_dir)

    @defer.inlineCallbacks
    def setup_config(self, path, data, remove=False):
        c = {}

        c['schedulers'] = [ForceScheduler(name="force", builderNames=["testy"])]

        f = BuildFactory()
        f.addStep(DownloadSecretsToWorker([(path, data)]))
        if remove:
            f.addStep(RemoveWorkerFileSecret([(path, data)]))

        c['builders'] = [BuilderConfig(name="testy", workernames=["local1"], factory=f)]

        yield self.setup_master(c)

    def get_homedir(self):
        path = os.path.expanduser('~')
        if path == '~':
            return None
        return path

    @parameterized.expand([
        ('simple', False, True),
        ('relative_to_home', True, True),
        ('simple_remove', False, True),
        ('relative_to_home_remove', True, True),
    ])
    @defer.inlineCallbacks
    def test_transfer_secrets(self, name, relative_to_home, remove):
        bb_path = self.temp_dir
        if relative_to_home:
            homedir = self.get_homedir()
            if homedir is None:
                raise SkipTest("Home directory is not known")
            try:
                bb_path = os.path.join('~', os.path.relpath(bb_path, homedir))
            except ValueError as e:
                raise SkipTest("Can't get relative path from home directory to test files") from e
            if not os.path.isdir(os.path.expanduser(bb_path)):
                raise SkipTest("Unknown error preparing test paths")

        path = os.path.join(bb_path, 'secret_path')
        data = 'some data'

        yield self.setup_config(path, data, remove=remove)

        yield self.doForceBuild()

        if remove:
            self.assertFalse(os.path.exists(path))
        else:
            self.assertTrue(os.path.isfile(path))
            with open(path, encoding='utf-8') as f:
                self.assertEqual(f.read(), data)


class DownloadSecretsBasePb(DownloadSecretsBase):
    proto = "pb"


class DownloadSecretsBaseMsgPack(DownloadSecretsBase):
    proto = "msgpack"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import json
import os
from unittest import mock

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import www
from buildbot.util import bytes2unicode
from buildbot.www import auth
from buildbot.www import config


class Utils(unittest.TestCase):
    def test_serialize_www_frontend_theme_to_css(self):
        self.maxDiff = None
        self.assertEqual(
            config.serialize_www_frontend_theme_to_css({}, indent=4),
            """\
--bb-sidebar-background-color: #30426a;
    --bb-sidebar-header-background-color: #273759;
    --bb-sidebar-header-text-color: #fff;
    --bb-sidebar-title-text-color: #627cb7;
    --bb-sidebar-footer-background-color: #273759;
    --bb-sidebar-button-text-color: #b2bfdc;
    --bb-sidebar-button-hover-background-color: #1b263d;
    --bb-sidebar-button-hover-text-color: #fff;
    --bb-sidebar-button-current-background-color: #273759;
    --bb-sidebar-button-current-text-color: #b2bfdc;
    --bb-sidebar-stripe-hover-color: #e99d1a;
    --bb-sidebar-stripe-current-color: #8c5e10;""",
        )


class TestConfigResource(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()

    @defer.inlineCallbacks
    def test_render(self):
        _auth = auth.NoAuth()
        _auth.maybeAutoLogin = mock.Mock()

        custom_versions = [['test compoent', '0.1.2'], ['test component 2', '0.2.1']]

        master = yield self.make_master(url='h:/a/b/', auth=_auth, versions=custom_versions)
        rsrc = config.ConfigResource(master)
        rsrc.reconfigResource(master.config)

        vjson = [list(v) for v in config.get_environment_versions()] + custom_versions

        res = yield self.render_resource(rsrc, b'/config')
        res = json.loads(bytes2unicode(res))
        exp = {
            "authz": {},
            "titleURL": "http://buildbot.net/",
            "versions": vjson,
            "title": "Buildbot",
            "auth": {"name": "NoAuth"},
            "user": {"anonymous": True},
            "buildbotURL": "h:/a/b/",
            "multiMaster": False,
            "port": None,
        }
        self.assertEqual(res, exp)


class IndexResourceTest(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()

    def get_react_base_path(self):
        path = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
        for _ in range(0, 4):
            path = os.path.dirname(path)
        return os.path.join(path, 'www/base')

    def find_matching_line(self, lines, match, start_i):
        for i in range(start_i, len(lines)):
            if match in lines[i]:
                return i
        return None

    def extract_config_json(self, res):
        lines = res.split('\n')

        first_line = self.find_matching_line(lines, '<script id="bb-config">', 0)
        if first_line is None:
            raise RuntimeError("Could not find first config line")
        first_line += 1

        last_line = self.find_matching_line(lines, '</script>', first_line)
        if last_line is None:
            raise RuntimeError("Could not find last config line")

        config_json = '\n'.join(lines[first_line:last_line])
        config_json = config_json.replace('window.buildbotFrontendConfig = ', '').strip()
        config_json = config_json.strip(';').strip()
        return json.loads(config_json)

    @parameterized.expand([
        ('anonymous_user', None, {'anonymous': True}),
        (
            'logged_in_user',
            {"name": 'me', "email": 'me@me.org'},
            {"email": "me@me.org", "name": "me"},
        ),
    ])
    @defer.inlineCallbacks
    def test_render(self, name, user_info, expected_user):
        _auth = auth.NoAuth()
        _auth.maybeAutoLogin = mock.Mock()

        custom_versions = [['test compoent', '0.1.2'], ['test component 2', '0.2.1']]

        master = yield self.make_master(
            url='h:/a/b/', auth=_auth, versions=custom_versions, plugins={}
        )
        if user_info is not None:
            master.session.user_info = user_info

        # IndexResource only uses static path to get index.html. In the source checkout
        # index.html resides not in www/base/public but in www/base. Thus
        # base path is sent to IndexResource.
        rsrc = config.IndexResource(master, self.get_react_base_path())
        rsrc.reconfigResource(master.config)

        vjson = [list(v) for v in config.get_environment_versions()] + custom_versions

        res = yield self.render_resource(rsrc, b'/')
        config_json = self.extract_config_json(bytes2unicode(res))

        _auth.maybeAutoLogin.assert_called_with(mock.ANY)
        exp = {
            "authz": {},
            "titleURL": "http://buildbot.net/",
            "versions": vjson,
            "title": "Buildbot",
            "auth": {"name": "NoAuth"},
            "user": expected_user,
            "buildbotURL": "h:/a/b/",
            "multiMaster": False,
            "port": None,
            "plugins": {},
        }
        self.assertEqual(config_json, exp)
